import logging
from string import Template
from typing import Dict

from rich.console import Console
from rich.live import Live
from rich.table import Table
from rich.text import Text

from .objects import RequestRecord


DEFAULT_COLUMNS = [
    {"header": "Method", "format": "[white]{method}", "style": "{"},
    {"header": "Path", "format": "[white bold]{path}", "style": "{"},
    {"header": "Status", "format": "{status_code}", "style": "{"},
    {"header": "Size", "format": "[white]{size}", "style": "{"},
    {
        "header": "Time",
        "format": "[white]{created}",
        "style": "{",
        "datefmt": "%H:%M:%S",
    },
]


class DjangoRequestHandler(logging.StreamHandler):
    """
    A logging handler that prints Django requests in a live-updating table.
    """

    live = None
    table = None
    console = None
    columns = None

    def __init__(self, *args, **kwargs):
        super().__init__()

        if "columns" in kwargs:
            self.columns = kwargs["columns"]
        else:
            self.columns = DEFAULT_COLUMNS

        if "console" in kwargs:
            self.console = kwargs["console"]

        if self.console is None:
            self.console = Console()

        if self.live is None:
            self.table = Table()
            self._add_column_headers()

            self.live = Live(self.table, auto_refresh=False)

    def _add_column_headers(self):
        """
        Add column headers for each column.
        """

        for column in self.columns:
            self.table.add_column(column["header"])

    def _add_row(self, request_record: RequestRecord) -> None:
        """
        Add a row to the live updating table that includes rendered markup for each column.
        """

        row_columns = []

        for column in self.columns:
            markup = self._get_markup(request_record, column)
            text = Text.from_markup(markup, style=request_record.text_style)

            row_columns.append(text)

        self.table.add_row(*row_columns)

    def _get_markup(self, request_record: RequestRecord, column: Dict) -> str:
        """
        Get markup based on the column's `format` and the `RequestRecord` object.
        """

        format = column.get("format", "")
        style = column.get("style", "%")
        date_format = column.get("datefmt")

        if style == "%":
            return format % request_record.get_dict(date_format=date_format)
        elif style == "{":
            return format.format(**request_record.get_dict(date_format=date_format))
        elif style == "$":
            template = Template(format)

            return template.substitute(
                **request_record.get_dict(date_format=date_format)
            )
        else:
            raise Exception(f"Unknown style: {style}")

    def emit(self, record: logging.LogRecord) -> None:
        """
        Writes the logging record. If it's a request, add the record to a live,
        updating table. If not, emit it like normal.
        """

        try:
            request_record = RequestRecord(record)

            if request_record.is_valid:
                if not request_record.is_loggable:
                    return

                self._add_row(request_record)

                self.live.start()
                self.live.refresh()

                self.flush()
            else:
                super().emit(record)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as ex:
            # self.handleError(record)
            print(ex)



import logging
import re
from dataclasses import dataclass
from datetime import datetime

from django.conf import settings


REQUEST_REGEX = (
    r"(GET|POST|PATCH|UPDATE|HEAD|PUT|DELETE|CONNECT|OPTIONS|TRACE)\s+([\S]+)\s+([\S]+)"
)


@dataclass
class RequestRecord:
    """
    Handles parsing the `logging.LogRecord` into a `RequestRecord` object.
    """

    method: str
    path: str
    status_code: str
    size: str
    created: datetime
    record: logging.LogRecord
    text_style: str = None
    is_valid: bool = False

    def __init__(self, record: logging.LogRecord):
        # Ignore any message from `django.server` that doesn't have 3 args.
        # TODO: Find a less hacky way to deal with this.
        if record.name != "django.server" or len(record.args) != 3:
            return

        self.record = record
        self.created = datetime.fromtimestamp(record.created)

        unparsed_request = record.args[0]

        # Example: GET /profile HTTP/1.1
        matches = re.match(REQUEST_REGEX, unparsed_request)

        if matches:
            self.is_valid = True

            self.method = matches.group(1)
            self.path = matches.group(2)

            self.status_code = record.args[1]
            self.size = record.args[2]

            self.text_style = self.get_text_style()

    def get_text_style(self):
        if self.status_code.startswith("2"):
            return "green"
        elif self.status_code.startswith("3"):
            return "yellow"
        elif self.status_code.startswith("4"):
            return "yellow"
        elif self.status_code.startswith("5"):
            return "red"

    @property
    def is_loggable(self):
        if settings.STATIC_URL and self.path.startswith(settings.STATIC_URL):
            return False

        if self.path == "/favicon.ico":
            return False

        return True

    def get_dict(self, date_format: str = None):
        created_str = str(self.created)

        if date_format:
            created_str = self.created.strftime(date_format)

        return {
            "created": created_str,
            "method": self.method,
            "path": self.path,
            "status_code": self.status_code,
            "size": self.size,
        }
